@@ -15,6 +15,7 @@ extern "C" {
 #include <libsoup/soup-auth-domain.h>
 #include <libsoup/soup-auth-domain-basic.h>
 #include <libsoup/soup-auth-domain-digest.h>
+#include <libsoup/soup-auth-manager.h>
 #include <libsoup/soup-cache.h>
 #include <libsoup/soup-content-decoder.h>
 #include <libsoup/soup-content-sniffer.h>
@@ -48,6 +49,7 @@ extern "C" {
 #include <libsoup/soup-tld.h>
 #include <libsoup/soup-uri.h>
 #include <libsoup/soup-value-utils.h>
+#include <libsoup/soup-version.h>
 #include <libsoup/soup-xmlrpc.h>
 
 #ifdef __cplusplus
